#!/usr/bin/perl -s

my $test_path;
my $cmd_path;
my $results_path;
my $max_processes;
my $repeatings;
my $bench_name;
my $zip_name;

sub is_nonempty ($) {
    return( defined $_[0] and length $_[0] != 0 );
}

if ( is_nonempty($localtestpath) ) {
    $test_path = $localtestpath;
} else {
	$test_path = '';
}

if ( is_nonempty($cmdpath) ) {
    $cmd_path = "$cmdpath\\";
} else {
	print "Warning: path to the command tool should be exported...\n";
	$cmd_path = '';
}

if ( is_nonempty($resultspath) ) {
    $results_path = "$resultspath\\";
} else {
	$results_path = '';
}

if ( is_nonempty($procs)) {
	$max_processes = $procs;
} else {
	$max_processes = 1;
}

if ( is_nonempty($rep)) {
	$repeatings = $rep;
} else {
	$repeatings = 10;
}

if (is_nonempty($benchname)) {
	# body...
	$bench_name = $benchname;
} else {
	$bench_name = "Benchmarks";
}

if (is_nonempty($zipname)) {
	# body...
	$zip_name = $zipname;
} else {
	$zip_name = "TestCases";
}


## This test script runs 'probcli' 

use strict;
use warnings;

#use File::Find::Rule;
use Cwd;
use POSIX qw/ceil/;
use Path::Class;

# Following packages should be installed for using this test script properly:
### Excel::Writer::XLSX, Statistics::Descriptive, Capture::Tiny
### I installed the perl packages with 'cpan' and I needed to run it in sudo mode
##### sudo cpan Excel::Writer::XLSX, sudo cpan Statistics::Descriptive, 
##### sudo cpan Capture::Tiny

use Capture::Tiny 'capture';

use Time::HiRes qw( sleep );
use Statistics::Descriptive;
use Excel::Writer::XLSX;

sub compute_and_add_excel_data {
	my($w,$w_stat,$r,$c,$c_stat,$use_ceil,@data) = @_; # read the parameters

	print "Calculating statistics of the following data: @data\n";
	my $stat = Statistics::Descriptive::Full->new();
	$stat->add_data(@data);
	my @stat_data;
	if ($use_ceil == 1) {
		@stat_data = (ceil($stat->geometric_mean()), ceil($stat->standard_deviation()), $stat->min(), $stat->max());
		$w->write( $r, $c, ceil($stat->geometric_mean()));
	} else {
		@stat_data = ($stat->geometric_mean(), $stat->standard_deviation(), $stat->min(), $stat->max());		
		$w->write( $r, $c, $stat->geometric_mean());
	}

	$c++;

	foreach my $d (@stat_data) {
		$w_stat->write( $r, $c_stat, $d );
		$c_stat++;
	}
	return ($c, $c_stat);
}

sub get_result {
	my($input) = @_;
	my $res;
	if ($input eq "") {
		$res = "No Error";
	} else { 
		if ($input =~ /.*\n..(.*)\n/) {
			$res = $1;
		} 
		else {
			$res = "Undefined error";
		}
	}
	return $res;
}

sub get_command {
	my($i) = @_;
	$i =~ m/(.*)=(.*)/;
	my $name = $1; my $cmd = $2;
	return ($name, $cmd);
}

sub get_files {
	my($dir) = @_;
	my @files;
	dir($dir)->recurse(callback => sub {
	   	my $file = shift;
	    if (!($file =~ m{^*rw_*}) & !($file =~ m{^*milnerscheduler*}) & $file =~ m{^*.csp$}) {
 	       push (@files, $file->absolute->stringify);
    }});
	return (@files);
}

sub write_info {
	my($w,$icmd,$format) = @_;
	my ($out, $err, $ret) = capture {
		system( $icmd );
	};
	$w->write(0, 1, 'Tool Info: ', $format);
	my $row = 1;
	foreach my $line (split m{\n}, $out ) {
		$w->write($row, 1, $line, $format);
		$row++;
	}
}

my $interrupt_catched = 0;
$SIG{'INT'} = 'INT_handler';

sub INT_handler {
	print "interrupt signal catched!\n";
	$interrupt_catched = 1;
}

my $cwd = getcwd();
my $filelist;

print "current cwd $cwd\n";
print "localtestpath: $test_path\n";

my $dir = "$cwd\\$test_path";

print "dir to be tested: $dir\n";
print "path to the main command: $cmd_path\n";

# my $excludeDirs = File::Find::Rule->directory
#                               ->name('') # Provide specific list of directories to *not* scan
#                               ->prune          # don't go into it
#                               ->discard;       # don't report it

# my $includeFiles = File::Find::Rule-> file
#                              ->name('*.eventb', '*.mch', '*.ref'); # search by file extensions

# my @files = File::Find::Rule->or( $excludeDirs, $includeFiles )
#                             ->in($dir);

my @files = &get_files($dir);

print "Files to be tested: @files\n";

# my $bench_name; my $zip_name;
if ($results_path eq '') {
 	$bench_name = "$bench_name.xlsx";
 	$zip_name = "$zip_name.zip";
} else {
	$bench_name = "$results_path\\$bench_name.xlsx";
	$zip_name = "$results_path\\$zip_name.zip";
}

my $workbook = Excel::Writer::XLSX->new( $bench_name );

# Add worksheets
my $worksheet_info = $workbook->add_worksheet('System and Tool Infos');
my $worksheet = $workbook->add_worksheet('Benchmarks');
my $worksheet_stat = $workbook->add_worksheet('Benchmarks\' Statistics');

# Adding and defining formats
my $format_col_items = $workbook->add_format();
$format_col_items->set_bold();
$format_col_items->set_color( 'black' );
$format_col_items->set_align( 'left' );

my $format_line_items = $workbook->add_format();
$format_line_items->set_bold();
$format_line_items->set_italic();
$format_line_items->set_color( 'black' );
$format_line_items->set_align( 'left' );

my $format_cell = $workbook->add_format();
$format_cell->set_align( 'center' );

## write info
#my @system_info = &get_system_info();
write_info($worksheet_info, ("\"$cmd_path" . "probcli.exe\"" . ' -version'), $format_col_items);

## Defining column data items
my $col; my $col_stat; my $row;

$row = $col = $col_stat = 0;

# a hash array for parsing output lines of the stdout of the command
# the elements "name => [pos, regexp]" of the hash array should be read as follows:
# 	 - name:   the name of the column in the table
#	 - pos:    the leftside position relative to the other elements (Note: the name should in the right column otherwise the putput data comes on different places)
#    - regexp: the regular expression that matches a line of the stdout of the executed tool
my %output_data = (
	'LTL Formula'               => [1, 'Checking LTL formula : (.*)'],
	'Atoms'                     => [2, 'LTL model checking: memory usage approx\. (.*) KiB, (\d+) atoms and (\d+) transitions generated'],
	'States'                    => [3, ''],
	'Rejected SCCs'             => [4, 'LTL model checking \(fairness\): (\d+) strongly connected components were rejected\.'],
    'Callbacks Time (Fairness)' => [5, ''],
	'Fairness Checking Time'    => [6, 'LTL model checking \(fairness\): total fairness checking time (\d+)ms, (\d+) callbacks needed (\d+)ms, netto (\d+)ms\.'],
    'Callbacks Time (Main)'     => [7, ''],
	'Total Time'                => [8, 'LTL model checking: total time (\d+)ms, (\d+) callbacks needed (\d+)ms, netto (\d+)ms\.']
	);

my @sorted_keys = sort {$output_data{$a}[0] <=> $output_data{$b}[0]} keys %output_data;

foreach my $item ('Model', 'Algorithm', 'Result', @sorted_keys, 'Overall Time') {
	$worksheet->write( 0, $col,  $item, $format_col_items );
	$col++;
}

my @stat_columns = ('Model', 'Algorithm', 'Result'); 

foreach my $key (@sorted_keys) {
    if ($key ne 'LTL Formula') {
        foreach my $stat ('geometric mean', 'standard deviation', 'min', 'max') {
            push(@stat_columns,"$key ($stat)");
        }
    }
}

push(@stat_columns, ('Overall Time (geometric mean)', 'Overall Time (standard deviation)', 'Overall Time (min)', 'Overall Time (max)'));

foreach my $item (@stat_columns) {
	$worksheet_stat->write( 0, $col_stat,  $item, $format_col_items );
	$col_stat++;
}

my @cmds = ('Fairness (ProB)=' . "\"$cmd_path" . "probcli.exe\"" . ' -ltlassertions --silent -csp-guide ');

### Using the fork manager for multiple processes
#use Parallel::ForkManager;
#use IPC::Shareable;

#my $pm = Parallel::ForkManager->new($max_processes);

#my $shmem = IPC::Shareable::SHM_BUFSIZ();

#print "shared memory size: $shmem\n";

# Created array in shared memmory
my %opts = (
  create    => 1,
  mode      => 0600,
  destroy   => 1
);
#my $key = 'mkey';
my @benchmark_results = (); 
#
#tie @benchmark_results, 'IPC::Shareable', $key, { %opts } or die "Failed to tie \@benchmark_results\n";

### Running the tests
foreach my $file (@files) {
	$col = $col_stat = 0;

	my @dirnames = (split m{\\}, $file);
	my $file_name = (@dirnames)[-2] . '\\' . (@dirnames)[-1];
	print "*************Testing: $file_name*************\n";
    my @file_benchmarks = ();

	# saving all benchmark results for the current file into one array
    my @cmd_data = ();

    # if (!$interrupt_catched) {
    # 	$pm->start and next;
    # }
	foreach my $i (@cmds) {
		my $wall_time;
		my @times = (); # for saving all measured times
		my $result; my $stdout; my $stderr; my $return;
		my @fairarr = (); my @sccarr = (); my @atomsarr = (); my @totalarr = ();
		my @statesarr = (); my @callbacksarr = (); my $formula; my @callbackfsarr = ();

		my ($name, $cmd) = &get_command($i);
		#print "commands: ($name, $cmd)\n";
		# getting time results
        my @test_benchmarks = ();

		for (0..($repeatings-1)) {
			sleep (1);
            my $file1 = "\"$file\"";
            my @filearr = (split m{\.}, $file);
            my $mchfile = "$filearr[0]" . ".mch";
			my $ex_cmd = $cmd . $file1 . " $mchfile";
			print "Executing command: $ex_cmd\n";
			($stdout, $stderr, $return) = capture {
				my $start_time = [Time::HiRes::gettimeofday()];
		  		system( $ex_cmd );
				$wall_time = Time::HiRes::tv_interval($start_time);
			};
			push (@times, $wall_time);
			print "\nTime spent on benchmark $wall_time\n\n";
			print "Parsing STDOUT...\n";
			my @lines = split(/\n/,$stdout);
			my $ana_set = 0;
			foreach my $line (@lines) {
				if ($line =~ /$output_data{'LTL Formula'}[1]/) {
					$formula = $1;
					# print "Formula checked: $1 $2\n"
				}
				if ($line =~ /$output_data{'Atoms'}[1]/) {
					# print "atoms checked $2\n";
					push (@atomsarr, $2);
				}
				if ($line =~ /$output_data{'Total Time'}[1]/) {
					push (@totalarr, $1/1000);
					push (@statesarr, $2);
					push (@callbacksarr, $3/1000);
				}				
				if ($line =~ /$output_data{'Rejected SCCs'}[1]/) {
					push (@sccarr, $1);
				}
				if ($line =~ /$output_data{'Fairness Checking Time'}[1]/) {
					push (@fairarr, $1/1000);
					push (@callbackfsarr, $2/1000);
				}
			}
			if ($interrupt_catched) {
				print "exit rep loop\n";
				last;
			}
		}

		$result = &get_result($stderr);

        push (@test_benchmarks, $name);
		push (@test_benchmarks, $result);
		push (@test_benchmarks, $formula);
		push (@test_benchmarks, \@atomsarr);
		push (@test_benchmarks, \@statesarr);
		push (@test_benchmarks, \@sccarr);
		push (@test_benchmarks, \@callbackfsarr);
		push (@test_benchmarks, \@fairarr);
		push (@test_benchmarks, \@callbacksarr);
		push (@test_benchmarks, \@totalarr);
		push (@test_benchmarks, \@times);

        push (@cmd_data, \@test_benchmarks);
		print "Test Result: $result\n";
		if ($interrupt_catched) {
			print "exit cmd loop\n";
			last;
		}
    }
    push (@file_benchmarks, $file_name);
    push (@file_benchmarks, \@cmd_data);
    
#    (tied @benchmark_results	)->shlock; # lock array for update
    push (@benchmark_results,\@file_benchmarks);
#    (tied @benchmark_results)->shunlock; # unlock the array

#    $pm->finish; # Terminates the child process
	if ($interrupt_catched) {
		print "exit files loop\n";
		last;
	}
}

#$pm->wait_all_children;

$SIG{'INT'} = 'DEFAULT'; # restore the default action

## Writing worksheets
print "Writing worksheets...";
foreach my $file (@benchmark_results) {
    
    $row++;
	$worksheet->write( $row, $col, $file->[0], $format_line_items );
	$worksheet_stat->write( $row, $col_stat, $file->[0], $format_line_items );

    my @stat_array = @{$file->[1]};
    
    foreach my $cmd (@stat_array) {

        $col = 1; $col_stat = 1;
        $worksheet->write( $row, $col, $cmd->[0]);
        $worksheet_stat->write( $row, $col_stat, $cmd->[0]);
        $col++; $col_stat++;
#        print "cmd name: $cmd->[0]\n";

        $worksheet->write( $row, $col, $cmd->[1], $format_cell);
        $worksheet_stat->write( $row, $col_stat, $cmd->[1], $format_cell);
        $col++; $col_stat++;
#        print "cmd result: $cmd->[1]\n";

        $worksheet->write( $row, $col, $cmd->[2], $format_cell);
        $worksheet_stat->write( $row, $col_stat, $cmd->[2], $format_cell);
        $col++; $col_stat++;

        print "Write statistics for test case \"$cmd->[0]\" of model \"$file->[0]\"\n";
        ($col, $col_stat) = &compute_and_add_excel_data($worksheet,$worksheet_stat,$row,$col,$col_stat,1,@{$cmd->[3]});
        ($col, $col_stat) = &compute_and_add_excel_data($worksheet,$worksheet_stat,$row,$col,$col_stat,0,@{$cmd->[4]});
        ($col, $col_stat) = &compute_and_add_excel_data($worksheet,$worksheet_stat,$row,$col,$col_stat,0,@{$cmd->[5]});
        ($col, $col_stat) = &compute_and_add_excel_data($worksheet,$worksheet_stat,$row,$col,$col_stat,0,@{$cmd->[6]});
        ($col, $col_stat) = &compute_and_add_excel_data($worksheet,$worksheet_stat,$row,$col,$col_stat,0,@{$cmd->[7]});
        ($col, $col_stat) = &compute_and_add_excel_data($worksheet,$worksheet_stat,$row,$col,$col_stat,0,@{$cmd->[8]});
        ($col, $col_stat) = &compute_and_add_excel_data($worksheet,$worksheet_stat,$row,$col,$col_stat,0,@{$cmd->[9]});
        ($col, $col_stat) = &compute_and_add_excel_data($worksheet,$worksheet_stat,$row,$col,$col_stat,0,@{$cmd->[10]});
#        print "stats: @{$cmd->[2]}\n";
        $row++;
    }
    $col = 0; $col_stat = 0;
}

$worksheet->write($row, 1, "Each test case has been executed $repeatings times.", $format_col_items);
$worksheet_stat->write($row, 1, "Each test case has been executed $repeatings times.", $format_col_items);

# clean up shared memory
# IPC::Shareable->clean_up;
# IPC::Shareable->clean_up_all;

## zipping the files
use Archive::Zip qw(:ERROR_CODES :CONSTANTS);
my $zip = Archive::Zip->new();

foreach my $file (@files) {
	my $fileName = 'TestCases' . '\\' . (split m{\\}, $file)[-2] . '\\' .(split m{\\}, $file)[-1];
	$zip->addFile( $file,  $fileName);
}
$zip->writeToFileNamed( $zip_name );
